import copy
import math
import uuid

from .mutex import mutex
from .topology import Topology


class Gossip:
    FANOUT = 5

    def __init__(self, transport, fanout=None, node_id=None):
        self.node_id = node_id or uuid.uuid4().hex
        self.transport = transport

        self.topology = Topology(self.node_id, self.transport.addr)

        self._fanout = fanout or self.FANOUT

    @property
    def fanout(self):
        return min(self._fanout, len(self.topology))

    @property
    def cycles(self):
        if self.fanout == 0:
            return 0

        if self.fanout == 1:
            return 1

        return math.ceil(math.log(len(self.topology), self.fanout))

    async def send(self, message, node_id):
        if node_id == self.node_id:
            raise ValueError("cannot send message to self")

        message["metadata"].setdefault("event", uuid.uuid4().hex)
        message["metadata"].setdefault("src", self.node_id)
        message["metadata"]["dst"] = node_id
        self.topology.set_route(message)

        addr = self.topology.get_addr(node_id)
        await self.transport.send(message, addr)

    async def send_ack(self, message):
        message = copy.deepcopy(message)
        if "ack" in message["metadata"]:
            return False

        message["metadata"].pop("event", None)
        message["metadata"].pop("src", None)

        message["metadata"]["ack"] = self.node_id
        await self.send(message, message["metadata"]["syn"])
        return True

    async def send_forward(self, message):
        if message["metadata"]["dst"] == self.node_id:
            return False
        else:
            await self.send(message, message["metadata"]["dst"])
            return True

    async def gossip(self, message):
        message["metadata"].pop("event", None)
        message["metadata"].pop("src", None)

        gossip_id = message["metadata"].setdefault("gossip", uuid.uuid4().hex)
        gossip_ignore = set([self.node_id])
        gossip_ignore.update([r[1] for r in message["metadata"].get("route", [])])

        @mutex(gossip_id, owner=self.gossip)
        async def multicast():
            cycle = 0
            while cycle < self.cycles:
                node_ids = self.topology.sample(self.fanout, ignore=gossip_ignore)
                for node_id in node_ids:
                    await self.send(message, node_id)
                gossip_ignore.update(node_ids)
                cycle += 1

        await multicast()

    async def recv(self):
        while True:
            message, addr = await self.transport.recv()
            message["metadata"]["route"][-1].append(list(addr))

            self.topology.set_route(message)
            node_ids = self.topology.update_routes(message["metadata"]["route"])

            # connect to new nodes
            for node_id in node_ids:
                await self.send({"metadata": {}}, node_id)

            # forward message to destination
            if await self.send_forward(message):
                continue

            # gossip message
            if "gossip" in message["metadata"]:
                await self.gossip(message)

            # acknowledge message
            if "syn" in message["metadata"]:
                await self.send_ack(message)

            yield message

    async def close(self):
        self.transport.close()



import asyncio
import collections
import fnmatch
import itertools

from .channel import Channel
from .errors import print_exception
from .gossip import Gossip


class Callback:
    def __init__(self, topic, func, loop: asyncio.AbstractEventLoop):
        self._loop = loop

        self.topic = topic
        self.func = func
        self.chan = Channel(loop=self._loop)

        self.task = self._loop.create_task(self())
        self.task.add_done_callback(print_exception)
        self._handler = None  # FIXME: refactor this to hooks

    async def __call__(self):
        while True:
            message = await self.chan.recv()
            result = await self.func(message)
            print(result)
            if result is None:
                continue

            if self._handler:
                await self._handler(message, result)

    async def cancel(self):
        await self.chan.close()
        self.task.cancel()


class Broker:
    TIMEOUT = 10

    def __init__(self, gossip: Gossip, loop: asyncio.AbstractEventLoop):
        self._loop = loop

        self.gossip = gossip
        self.callbacks = collections.defaultdict(list)

    async def listen(self):
        """Connect to the gossip network and start receiving messages."""
        async for message in self.gossip.recv():
            # FIXME: make messages idempotent (prevent duplicate processing)

            if "topic" not in message["metadata"]:
                continue

            topics = list(self.callbacks.keys())

            for topic in topics:
                if fnmatch.fnmatch(message["metadata"]["topic"], topic):
                    for callback in self.callbacks[topic]:
                        await callback.chan.send(message)

            for topic in topics:
                if len(self.callbacks[topic]) == 0:
                    del self.callbacks[topic]

    async def close(self):
        """Disconnect from the gossip network and stop receiving messages."""
        callbacks = [cb for cb in itertools.chain(*self.callbacks.values())]
        await asyncio.gather(*[cb.cancel() for cb in callbacks], return_exceptions=True)
        await self.gossip.close()

    def subscribe(self, topic, callback):
        """Subscribe to a topic and register a callback."""
        callback = Callback(topic, callback, loop=self._loop)
        self.callbacks[topic].append(callback)
        return callback

    async def unsubscribe(self, callback):
        """Unsubscribe from a topic and unregister a callback."""
        await callback.cancel()
        self.callbacks[callback.topic].remove(callback)

    async def publish(self, topic, message, node_ids=None):
        """Publish a message to a topic."""
        # FIXME: make messages idempotent (prevent duplicate processing)
        # FIXME: allow sending to specific nodes
        message["metadata"]["topic"] = topic
        if node_ids:
            for node_id in node_ids:
                if node_id in self.gossip.topology:
                    await self.gossip.send(message, node_id)
                else:
                    raise ValueError(f"Unknown node: {node_id}")
        else:
            await self.gossip.gossip(message)

        if "syn" in message["metadata"]:
            return self._recv(topic, node_ids=node_ids)

    async def _recv(self, topic, node_ids=None):
        chan = Channel(loop=self._loop)
        callback = self.subscribe(topic, chan.send)

        acks = set()
        if node_ids:
            acks.update(node_ids)

        try:
            async with asyncio.timeout(self.TIMEOUT):
                while True:
                    message = await chan.recv()
                    if "ack" in message["metadata"]:
                        acks.add(message["metadata"]["ack"])
                        yield message

                    elif message["metadata"]["src"] in acks:
                        acks.remove(message["metadata"]["src"])
                        yield message

                    else:
                        raise ValueError(f"Unknown message: {message}")

        except asyncio.TimeoutError:
            pass
        finally:
            await self.unsubscribe(callback)
            await chan.close()



"""
Usage:
    GOSSIP_SEEDS=4aed4f3c6900462baa1413fb7ef4f814@127.0.0.1:58295 python examples/members.py
"""
import asyncio
import json

import aiogossip

peer = aiogossip.Peer()


@peer.subscribe("*")
async def on_message(message):
    print("---")
    print("receive: ", json.dumps(message))


async def main():
    while True:
        await asyncio.sleep(10)
        print("---")
        print("members:")
        for node in peer.nodes:
            print("-", node)
        print()


if __name__ == "__main__":
    peer.connect()
    peer.run_forever(main=main)



import asyncio
import uuid

from . import config
from .broker import Broker
from .errors import print_exception
from .gossip import Gossip
from .transport import Transport


class Peer:
    def __init__(
        self,
        loop: asyncio.AbstractEventLoop = None,
        host="0.0.0.0",
        port=0,
        fanout=None,
        node_id=None,
    ):
        self._loop = loop or asyncio.get_event_loop()

        self.node_id = node_id or uuid.uuid4().hex
        # FIXME: should be lazy
        self.transport = Transport((host, port), loop=self._loop)
        self.gossip = Gossip(self.transport, fanout=fanout, node_id=self.node_id)
        self.broker = Broker(self.gossip, loop=self._loop)

        self.task = self._loop.create_task(self.broker.listen())
        self.task.add_done_callback(print_exception)

    @property
    def node(self):
        return self.gossip.topology.node

    @property
    def DSN(self):
        return "{}@{}:{}".format(self.node["node_id"], *self.node["node_addr"])

    @property
    def nodes(self):
        return list(self.gossip.topology)

    async def _connect(self):
        topic = "connect:{}".format(uuid.uuid4().hex)
        message = {"metadata": {}}

        response = await self.publish(topic, message, syn=True)
        async for r in response:
            pass

    def connect(self, seeds=config.GOSSIP_SEEDS):
        if not seeds:
            return

        nodes = []

        if isinstance(seeds, str):
            seeds = seeds.split(",")
            for seed in seeds:
                node_id, addr = seed.split("@")
                host, port = addr.split(":")
                nodes.append({"node_id": node_id, "node_addr": (host, int(port))})

        elif isinstance(seeds, list):
            nodes = seeds

        self.gossip.topology.add(nodes)
        task = self._loop.create_task(self._connect())
        task.add_done_callback(print_exception)

    async def disconnect(self):
        await self.broker.close()
        self.task.cancel()
        await asyncio.gather(self.task, return_exceptions=True)

    async def publish(self, topic, message, peers=None, syn=False):
        topic = topic.replace("{uuid}", uuid.uuid4().hex)

        if syn:
            message["metadata"]["syn"] = self.node_id
        return await self.broker.publish(topic, message, node_ids=peers)

    def subscribe(self, topic):
        def decorator(callback):
            callback = self.broker.subscribe(topic, callback)
            return callback

        return decorator

    async def request(self, topic, message, peers=None, timeout=5):
        topic = "request:{}:{}".format(topic, uuid.uuid4().hex)
        response = await self.publish(topic, message, peers=peers, syn=True)
        return response

    def response(self, topic):
        topic = "request:{}:*".format(topic)

        async def responder(message, r):
            await self.publish(
                message["metadata"]["topic"], r, peers=[message["metadata"]["syn"]]
            )

        def decorator(callback):
            callback = self.broker.subscribe(topic, callback)
            callback._handler = responder
            return callback

        return decorator

    def run_forever(self, main=None):  # pragma: no cover
        if main:
            self._loop.run_until_complete(main())

        try:
            self._loop.run_forever()
        except KeyboardInterrupt:
            pass
        finally:
            self._loop.run_until_complete(self.disconnect())
